@@ -127,7 +127,7 @@ static tcn_ssl_conn_t *ssl_create(JNIEnv *env, tcn_ssl_ctxt_t *ctx, apr_pool_t *
     }
     if ((ssl = SSL_new(ctx->ctx)) == NULL) {
         char err[256];
-        ERR_error_string(ERR_get_error(), err);
+        ERR_error_string(SSL_ERR_get(), err);
         tcn_Throw(env, "SSL_new failed (%s)", err);
         con = NULL;
         return NULL;
@@ -320,7 +320,7 @@ TCN_IMPLEMENT_CALL(jint, SSLSocket, handshake)(TCN_STDARGS, jlong sock)
 
     apr_socket_timeout_get(con->sock, &timeout);
     while (!SSL_is_init_finished(con->ssl)) {
-        ERR_clear_error();
+        SSL_ERR_clear();
         if ((s = SSL_do_handshake(con->ssl)) <= 0) {
             if (!con->ssl)
                 return APR_ENOTSOCK;
@@ -339,8 +339,11 @@ TCN_IMPLEMENT_CALL(jint, SSLSocket, handshake)(TCN_STDARGS, jlong sock)
                     }
                 break;
                 case SSL_ERROR_SYSCALL:
+                    if (APR_STATUS_IS_EAGAIN(rv)) {
+                        return APR_EAGAIN;
+                    }
 #if !defined(_WIN32)
-                    if (APR_STATUS_IS_EINTR(rv)) {
+                    else if (APR_STATUS_IS_EINTR(rv)) {
                         /* Interrupted by signal */
                         continue;
                     }
@@ -362,13 +365,12 @@ TCN_IMPLEMENT_CALL(jint, SSLSocket, handshake)(TCN_STDARGS, jlong sock)
         * Check for failed client authentication
         */
         if (con->ctx->verify_mode != SSL_VERIFY_NONE &&
-	    (vr = SSL_get_verify_result(con->ssl)) != X509_V_OK) {
+                (vr = SSL_get_verify_result(con->ssl)) != X509_V_OK) {
 
             if (SSL_VERIFY_ERROR_IS_OPTIONAL(vr) &&
-                con->ctx->verify_mode == SSL_CVERIFY_OPTIONAL_NO_CA) {
+                    con->ctx->verify_mode == SSL_CVERIFY_OPTIONAL_NO_CA) {
                 /* TODO: Log optionalNoCA */
-            }
-            else {
+            } else {
                 /* TODO: Log SSL client authentication failed */
                 con->shutdown_type = SSL_SHUTDOWN_TYPE_UNCLEAN;
                 /* TODO: Figure out the correct return value */
@@ -403,7 +405,7 @@ ssl_socket_recv(apr_socket_t *sock, char *buf, apr_size_t *len)
     }
     apr_socket_timeout_get(con->sock, &timeout);
     for (;;) {
-        ERR_clear_error();
+        SSL_ERR_clear();
         if ((s = SSL_read(con->ssl, buf, rd)) <= 0) {
             if (!con->ssl)
                 return APR_ENOTSOCK;
@@ -427,6 +429,9 @@ ssl_socket_recv(apr_socket_t *sock, char *buf, apr_size_t *len)
                         con->shutdown_type = SSL_SHUTDOWN_TYPE_STANDARD;
                         return APR_EOF;
                     }
+                    else if (APR_STATUS_IS_EAGAIN(rv)) {
+                        return APR_EAGAIN;
+                    }
 #if !defined(_WIN32)
                     else if (APR_STATUS_IS_EINTR(rv)) {
                         /* Interrupted by signal
@@ -482,7 +487,7 @@ ssl_socket_send(apr_socket_t *sock, const char *buf,
     }
     apr_socket_timeout_get(con->sock, &timeout);
     for (;;) {
-        ERR_clear_error();
+        SSL_ERR_clear();
         if ((s = SSL_write(con->ssl, buf, wr)) <= 0) {
             if (!con->ssl)
                 return APR_ENOTSOCK;
@@ -611,52 +616,122 @@ TCN_IMPLEMENT_CALL(jint, SSLSocket, attach)(TCN_STDARGS, jlong ctx,
     return APR_SUCCESS;
 }
 
+static int ssl_do_renegotiate(tcn_ssl_conn_t *con, int use_pha)
+{
+    int retVal;
+    int error = 0;
+    char peekbuf[1];
+    apr_interval_time_t timeout;
+
+    apr_socket_timeout_get(con->sock, &timeout);
+
+    /* Trigger reading of the certs from the client.
+     * Peeking 0 bytes actually works.
+     * Before TLS 1.3 this will result in a renegotiation.
+     * for TLS 1.3 in PHA.
+     * See: http://marc.info/?t=145493359200002&r=1&w=2
+     *
+     * This will normally return SSL_ERROR_WANT_READ whether the renegotiation
+     * has been completed or not. Afterwards, need to determine if I/O needs to
+     * be triggered or not.
+     */
+    retVal = SSL_peek(con->ssl, peekbuf, 0);
+    if (retVal < 1) {
+        error = SSL_get_error(con->ssl, retVal);
+    }
+
+    // If the certs have not been received, then need to wait for I/O
+    while ((use_pha && con->pha_state == PHA_STARTED) || (!use_pha && SSL_renegotiate_pending(con->ssl))) {
+        // SSL_ERROR_WANT_READ is expected. Anything else is an error.
+        if (error == SSL_ERROR_WANT_READ) {
+            retVal = wait_for_io_or_timeout(con, error, timeout);
+            /*
+             * Since this is blocking I/O, anything other than APR_SUCCESS is an
+             * error.
+             */
+            if (retVal != APR_SUCCESS) {
+                con->shutdown_type = SSL_SHUTDOWN_TYPE_UNCLEAN;
+                return retVal;
+            }
+        } else {
+            return APR_EGENERAL;
+        }
+
+        // Re-try SSL_peek after I/O
+        retVal = SSL_peek(con->ssl, peekbuf, 0);
+        if (retVal < 1) {
+            error = SSL_get_error(con->ssl, retVal);
+        } else {
+            /*
+             * Reset error to handle case where SSL_Peek returns 0 but
+             * the pha resp. renegotiation state has not changed.
+             * This will trigger an error to be returned.
+             */
+            error = 0;
+        }
+    }
+    return APR_SUCCESS;
+}
+
 TCN_IMPLEMENT_CALL(jint, SSLSocket, renegotiate)(TCN_STDARGS,
                                                  jlong sock)
 {
     tcn_socket_t *s   = J2P(sock, tcn_socket_t *);
     tcn_ssl_conn_t *con;
     int retVal;
-    char peekbuf[1];
+#if defined(SSL_OP_NO_TLSv1_3)
+    const SSL_SESSION *session;
+#endif
 
     UNREFERENCED_STDARGS;
     TCN_ASSERT(sock != 0);
     con = (tcn_ssl_conn_t *)s->opaque;
 
-    /* Sequence to renegotiate is
-     *  SSL_renegotiate()
-     *  SSL_do_handshake()
-     *  ssl->state = SSL_ST_ACCEPT
-     *  SSL_do_handshake()
-     */
+#if defined(SSL_OP_NO_TLSv1_3)
+    session  = SSL_get_session(con->ssl);
+    if (SSL_SESSION_get_protocol_version(session) == TLS1_3_VERSION) {
+        // TLS 1.3 renegotiation
+        retVal = SSL_verify_client_post_handshake(con->ssl);
+        if (retVal <= 0) {
+            return APR_EGENERAL;
+        }
 
-    /* Toggle the renegotiation state to allow the new
-     * handshake to proceed.
-     */
-    con->reneg_state = RENEG_ALLOW;
-    retVal = SSL_renegotiate(con->ssl);
-    if (retVal <= 0)
-        return APR_EGENERAL;
+        con->pha_state = PHA_STARTED;
 
-    retVal = SSL_do_handshake(con->ssl);
-    if (retVal <= 0)
-        return APR_EGENERAL;
-    if (!SSL_is_init_finished(con->ssl)) {
-        return APR_EGENERAL;
+        // Need to trigger a write operation to sent the cert request to the
+        // client. As per OpenSSL docs, use SSL_do_handshake() for this.
+        retVal = SSL_do_handshake(con->ssl);
+        if (retVal <= 0) {
+            return APR_EGENERAL;
+        }
+        retVal = ssl_do_renegotiate(con, 1);
+        if (retVal != APR_SUCCESS) {
+            return retVal;
+        }
+
+    } else {
+#endif
+        // TLS 1.2 and earlier renegotiation
+
+        /* Toggle the renegotiation state to allow the new
+         * handshake to proceed.
+         */
+        con->reneg_state = RENEG_ALLOW;
+
+        // Schedule a renegotiation request
+        retVal = SSL_renegotiate(con->ssl);
+        if (retVal <= 0) {
+            return APR_EGENERAL;
+        }
+        retVal = ssl_do_renegotiate(con, 0);
+        if (retVal != APR_SUCCESS) {
+            return retVal;
+        }
+
+        con->reneg_state = RENEG_REJECT;
+#if defined(SSL_OP_NO_TLSv1_3)
     }
-
-    /* Need to trigger renegotiation handshake by reading.
-     * Peeking 0 bytes actually works.
-     * See: http://marc.info/?t=145493359200002&r=1&w=2
-     */
-    SSL_peek(con->ssl, peekbuf, 0);
-
-    con->reneg_state = RENEG_REJECT;
-
-    if (!SSL_is_init_finished(con->ssl)) {
-        return APR_EGENERAL;
-    }
-
+#endif
     return APR_SUCCESS;
 }
 
